<h2>AES Ciphertext Collisions Anomaly</h2>
<i>by Javantea</i><br />
<i>April 8-10, 2015</i><br />

<style>
<!--
blockquote { font-family:"Helvetica Neue",Helvetica,Arial,sans-serif; }
-->
</style>

<h3>Introduction</h3>
<!--<blockquote>
</blockquote>-->

<a href="https://www.altsci.com/concepts/aes/">Permalink</a><br />
<a href="keystream_dupe-0.3.tgz">keystream_dupe-0.3.tgz</a> <a href="keystream_dupe-0.3.tgz.asc">[sig]</a><br />
<a href="keystream_dupe-0.2.tgz" style="color:#808080;">keystream_dupe-0.2.tgz</a> <a href="keystream_dupe-0.2.tgz.asc" style="color:#808080;">[sig]</a><br />
<a href="keystream_dupe-0.1.tgz" style="color:#808080;">keystream_dupe-0.1.tgz</a> <a href="keystream_dupe-0.1.tgz.asc" style="color:#808080;">[sig]</a><br />

<p>
In this very basic cryptography exercise, I have written a simple test of the quality of a cipher. For RC4 and stream ciphers, we can encrypt <code>\x00\x00\x00\x00</code> to get the first four bytes of the keystream. I do this for the first 1048576 keys (assuming big endian and 64-bit keys) with RC4. Then I find out how many random keys I have to try before I find the same first four keystream bytes. I do this 1024 times. The data shows that this is around 4 million keys.
</p>
<p>
For block ciphers like AES, we have to do it slightly differently, but the concept is the exact same. I encrypt "GET / HTTP/1.1\r\n" which happens to be 16 bytes, the exactly correct size to fit in a single block of AES plaintext. I store the first four bytes of the ciphertext for the first 1048576 keys (same assumption as above but 128-bit keys). Then I do the same with random keys and I compare the first four bytes of the ciphertext against the first four bytes of the 1048576 partial ciphertexts. I find out how many random keys I have to try before I find the same first four ciphertext bytes. I do this 1024 times. The data shows that this is around 3 million keys. As you can clearly see, this is far smaller than RC4 (which is known to be vulnerable to many attacks).
</p>

<p>
This appears to be a weakness in AES. My lack of knowledge of cryptography however is making this educated guess poorly backed by science. I hope that a cryptographer can look at the source code and the methodology and tell me why I am seeing this result.
</p>

<h3>Data</h3>
<h4>Attempts required to collide 1024 ciphertexts/keystreams</h4>
<pre>
AES:
1024 keys found in 3289822 attempts
1024 keys found in 2919389 attempts
1024 keys found in 3160136 attempts
1024 keys found in 3169387 attempts

RC4:
1024 keys found in 4081981 attempts
1024 keys found in 4038271 attempts
1024 keys found in 4058535 attempts
1024 keys found in 4121532 attempts

</pre>
<h4>Output of getKeyStreams(), getCiphertexts()</h4>
<pre>
AES:
len(tons_of_ciphertexts_set): 1048451 1.000
len(tons_of_ciphertexts_set): 1048451 1.000
len(tons_of_ciphertexts_set): 1048451 1.000
len(tons_of_ciphertexts_set): 1048451 1.000

RC4:
len(tons_of_keystreams_set): 1047777 0.999
len(tons_of_keystreams_set): 1047777 0.999
len(tons_of_keystreams_set): 1047777 0.999
len(tons_of_keystreams_set): 1047777 0.999

</pre>

<h4>Fixed version of RC4</h4>
<pre>
RC4 (with the first 1024 bytes discarded):
len(tons_of_keystreams_set): 1048448 1.000
len(tons_of_keystreams_set): 1048448 1.000
1024 keys found in 4102351 attempts
1024 keys found in 4120705 attempts
</pre>

<h3>Analysis</h3>
<p>
Cryptographers on Twitter have come to my aid on this project. <a href="https://twitter.com/jedisct1">Frank Denis ‏@jedisct1</a> correctly identified the problem as <a href="https://en.wikipedia.org/wiki/Birthday_problem">the Birthday Problem</a>. This does not solve the issue I am seeing here. Continuing on, more insightful comments come from <a href="https://twitter.com/marshray">Marsh Ray</a> (a respected security and cryptography researcher I know here in the area) who says:
</p>
<blockquote>
I suspect it's actually a weakness of RC4 you're seeing. It tends to have pairs of same values<br /><br />
But your plaintext has no repeating values, so when XORd will have a smaller than chance probability of being \x00\x00\x00\x00
</blockquote>
<p>
This is most likely the correct answer. RC4 has properties that a perfect pseudorandom number generator (PRNG) does not have. The question now becomes why exactly does <code>AES("GET / HTTP/1.1\r\n", random_key)[:4]</code> collide sooner than a bad PRNG? The expectation is that AES is a nearly perfect PRNG. If this is not true, then AES-CTR is at least partially broken.
</p>
<p>
<a href="https://twitter.com/DefuseSec">Taylor Hornby (@DefuseSec)</a> correctly identified a bug in my code that did not affect the functioning of the code (it was equivalent to a nop). I have fixed that line of code and released a new version. The old version is also available to those who wish to compare.
</p>
<p>
<a href="https://twitter.com/eqe">Andy Isaacson (@eqe)</a> spent several posts trying to explain concisely the problem:
</p>
<blockquote>
with AES, you've computed 2^20 random 32 bit values. Check how many different values you got - should be about 1048449.<br /><br />
now, you're computing more random 32-bit values and checking if they're in the initial list. Each random has a 1 in 2^12 chance.<br /><br />
this is straight probability -- how long until you hit 50% probability of collision? It's simply pow(1-(1.0/(1<<12)), 2900)=0.49 !<br /><br />
(well not quite 1 in 2^12, but rather a 1048449 in 2^32 chance;
0.0244111% versus
0.0244140%. Close enough.)<br /><br />
I suspect your RC4 initial set is smaller since RC4 is biased, decreasing the probability of your random sequence colliding.<br /><br />
but I don't have an RC4 implementation in this crypto library, so you'll have to check for yourself. :)<br /><br />
</blockquote>
<p>
This is correct, the RC4 initial set is substantially smaller than the AES initial set because RC4 collides ciphertexts in the getKeyStreams function while AES does not collide Ciphertexts in the getCiphertexts function. Excellent detective work without a functioning implementation of RC4 Andy.
</p>
<p>
Andy's answer doesn't actually solve the discrepancy between AES and RC4. When RC4 is changed so that doesn't collide 674 times in getKeyStreams, RC4 still takes 4102351 attempts to collide 1024 times. AES still takes ~3 million attempts.
</p>
<p>
Analysis will continue until an answer is found. Currently no explanation for the phenomena exists.
</p>

<h3>Conclusion</h3>
<p>
<strike>Andy Isaacson (@eqe) successfully determined that when conducting a birthday attack, the number of items being collided against plays a very big role on how many collisions occur per N attempts. The data that I failed to put into the original is now in the data section.</strike>
</p>

<p>
<strike>This short foray into cryptography shows no weakness in AES. Instead it shows an easily discovered weakness in RC4. By generating 1048451 4-byte keystreams, 674 collisions occur. Modern implementations of RC4 throw away the first 1024 bytes of the keystream, so this is an unfair assessment, so version 0.3 will test this modified version of RC4.</strike>
</p>

<p>
This mystery continues unsolved.
</p>
